# frozen_string_literal: true

require 'singleton'

class Money
  attr_reader :amount, :currency

  def self.conversion_rates(*args)
    base         = args[0]
    values       = { base => 1.0 }.merge(args[1])
    Rate.instance.define_attrs(base, values)
  end

  def initialize(*args)
    @amount, @currency = args
  end

  def inspect
    "#{format('%.2f', amount)} #{currency}"
  end

  def convert_to(new_currency)
    value = Rate.instance.convert_to(
      amount:           amount,
      origin_currency:  currency,
      new_currency:     new_currency
    )

    Money.new(value, new_currency)
  end

  def +(other)
    value = Rate.instance.convert_to(
      amount:           other.amount,
      origin_currency:  other.currency,
      new_currency:     currency
    )

    Money.new(value + amount, currency)
  end

  def -(other)
    value = Rate.instance.convert_to(
      amount:           other.amount,
      origin_currency:  other.currency,
      new_currency:     currency
    )

    Money.new(amount - value, currency)
  end

  def /(other)
    Money.new(amount / other, currency)
  end

  def *(other)
    Money.new(amount * other, currency)
  end

  def ==(other)
    other.convert_to(currency).amount.round(2) == amount.round(2)
  end

  def >(other)
    amount.round(2) > other.convert_to(currency).amount.round(2)
  end

  def <(other)
    amount.round(2) < other.convert_to(currency).amount.round(2)
  end

  class Rate
    include Singleton

    def define_attrs(*args)
      @base, @values = args.flatten
    end

    def convert_to(amount:, origin_currency:, new_currency:)
      base_value = move_currency_to_base(amount, origin_currency)

      base_value * values[new_currency]
    end

    private

    attr_reader :base, :values

    def move_currency_to_base(amount, currency)
      return amount if currency == base

      amount / values[currency]
    end
  end

  private_constant :Rate
end
